#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/types.h>
#include "search.h"
#include "config.h"
#include "charset.h"
#include "common.h"
#include "udm_syslog.h"
#include "dbms.h"

#define T_UNKNOWN	0
#define T_TOP		1
#define T_BOT		2
#define T_RESTOP	3
#define T_RESBOT	4
#define T_RESULT	5
#define T_ERROR		6
#define T_NOTFOUND	7
#define T_CLONE		8
#define T_NAVIGATOR	9
#define T_NAVLEFT	10
#define T_NAVBAR0	11
#define T_NAVBAR1	12
#define T_NAVRIGHT	13

int Randoms[MAXRANDOM];

#define MAXTEMPLATE 1024
typedef struct template_struct{
	int where;
	char *str;
} TEMPLATE;
static TEMPLATE Template[MAXTEMPLATE];
static int ntempl=0;


/* Template variables */
static char error[BUFSIZ]="";
static char wordinfo[STRSIZ]="";
static char navigator[STRSIZ]="";
static char href[STRSIZ]="";
static char fullhref[STRSIZ]="";
static char *query_form_escaped=0;
static char *crc=0;
static char *self=0;
static char *title=0;
static char *text=0;
static int  size=0;
static int  rating=0;
static char *descr=0;
static char *keywords=0;
static int  url_id,docnum,first=0,last=0,found=0;
static int num_pages=0, navpage=0, ps=0, pps=10,np=0,is_next=0;


char* htmlspecialchars(char *str){
char *pos, *p;
int len;
	if(!str)return(NULL);
	len = strlen(str);
	for(pos = str; *pos; pos++){
		switch (*pos){
			case '&': len+=4; break;
			case '"': len+=5; break;
			case '<':
			case '>': len+=3; break;
		}
	}
	len;

	p = (char*) malloc(len+1);*p = 0;
	for(pos=p; *str; str++) {
	switch (*str){
		case '&': strcat(p, "&amp;"); pos+=5; break;
		case '"': strcat(p, "&quot;");pos+=6; break;
		case '<': strcat(p, "&lt;");pos+=4; break;
		case '>': strcat(p, "&gt;");pos+=4; break;
		default:
			*pos = *str;
			pos++;*pos=0;
		}
	}
	return p;
}

char * Unescape(char *d,char *s){
int hi,lo=0;
char *dd;
	if((d==NULL)||(s==NULL))return(0);
	dd=d;
	while(*s){
		if(*s=='%'){
			if(strchr("0123456789",*(++s))) hi=*s-'0';
			else hi=*s-'A'+10;
			if(strchr("0123456789",*(++s))) lo=*s-'0';
			else lo=*s-'A'+10;
			*d=hi*16+lo;
		}else
		if(*s=='+'){
			*d=' ';
		}else{
			*d=*s;
		}
		s++; d++;
	}
	*d=0;return(dd);
}

char *Escape(char *d,char *s,char what){
int hi,lo=0;
char *dd;
	if((d==NULL)||(s==NULL))return(0);
	dd=d;
	while(*s){
		if(*s==what){
			if(what=='"') {strcpy(d,"&quot;"); d+=5;}else
			if(what==' ') {*d='+';}
		}else{
			*d=*s;
		}
		s++; d++;
	}
	*d=0;
	return(dd);
}

char * EscapeURL(char *d,char *s){
char *dd;
	if((d==NULL)||(s==NULL))return(0);
	dd=d;
	while(*s){
		if(strchr("<>+[](){}/?#'\"\\;,",*s)){
			sprintf(d,"%%%x",(int)*s);
			*(d+1)=toupper(*(d+1));
			*(d+2)=toupper(*(d+2));
			d+=2;
		}else
		if(*s==' '){
			*d='+';
		}else{
			*d=*s;
		}
		s++;d++;
	}
	*d=0;
	return(dd);
}



int PrintTemplate(int where,char *url,char *content_type,char *last_modified)
{
int i,rnum,tmp;
char *s;
DOCUMENT *r=NULL;

	for(i=0;i<ntempl;i++){
		if(Template[i].where!=where)continue;
		s=Template[i].str;
		while(*s){
			if((*s=='\\')&&(*(s+1)=='$')){printf("$");s+=2;continue;}
			if(*s!='$'){printf("%c",*s);fflush(stdout);s++;continue;}
			s++;
			switch(*s){
			case 'A': printf("%s",self);break;
			case 'Q': printf("%s",query_form_escaped?query_form_escaped:"");break;
			case 'E': printf("%s",error);break;
			case 'W': printf("%s",wordinfo);break;
			case 'V': {
				if(num_pages>1)PrintTemplate(T_NAVIGATOR,"","","");
				break;
			}
			case 'f': printf("%d",first);break;
			case 'l': printf("%d",last);break;
			case 't': printf("%d",found);break;
			case 'r': {
				s++;
				rnum=atoi(s);
				if((rnum>=0)&&(rnum<MAXRANDOM))
					printf("%d",Randoms[rnum]);
				while((*s)&&(isdigit(*(s+1))))
					s++;
				break;
			}

			case 'D':
				s++;
				switch(*s){
				case 'U':printf("%s",url?url:"");break;
				case 'T':printf("%s",title?title:"");break;
				case 'C':printf("%s",content_type?content_type:"");break;
				case 'M':printf("%s",last_modified?last_modified:"");break;
				case 'X':printf("%s",text?text:"");break;
				case 'R':printf("%d",rating);break;
				case 'S':printf("%d",size);break;
				case 'N':printf("%d",docnum);break;
				case 'D':printf("%s",descr);break;
				case 'K':printf("%s",keywords);break;
				default :printf("%c",*s);
				}
				break;

			case 'C':
				s++;
				switch(*s){
				case 'L':
				if((crc)&&(*crc)&&(where==T_RESULT)){
					if(r=CloneList(crc)){
						while(r->url_id){
							if(r->url_id!=url_id)
							PrintTemplate(T_CLONE,r->url,r->content_type,r->last_modified);
							r++;
						}
					}
				}
				break;
				default : printf("%c",*s);
				}
				break;
			case 'N':
				s++;
				switch(*s){
					case 'L':
					if (np > 0) {
						if(np>1)sprintf(fullhref,"%s&np=%d",href,np-1);
						else strcpy(fullhref,href);
						PrintTemplate(T_NAVLEFT,"","","");
					}
					break;
					case 'B':
					for(navpage = 1; navpage <= num_pages;navpage++){
						tmp = np + 1 - pps/2;
						if (tmp < 1) tmp = 1;
						if (tmp > num_pages - pps + 1)
						tmp = num_pages - pps + 1;
						if (navpage >= tmp && navpage <(tmp+pps)){
							if(navpage>1)sprintf(fullhref,"%s&np=%d",href,navpage-1);
							else strcpy(fullhref,href);
							if (np == navpage-1)
							PrintTemplate(T_NAVBAR0,"","","");
						else
							PrintTemplate(T_NAVBAR1,"","","");
						}
					}
				break;
					case 'R':
					if (is_next) {
						sprintf(fullhref,"%s&np=%d",href,np+1);
						PrintTemplate(T_NAVRIGHT,"","","");
					}
					break;
					case 'H': printf("%s",fullhref);break;
					case 'P': printf("%d",navpage);break;
				}
				break;
			default: printf("%c",*s);
			}
			s++;
		}
	}
}



main(int argc, char **argv) {
FILE *file;
float frand;
time_t tclock;
char wordchar[256]="";
char str[STRSIZ]="";
char str1[STRSIZ]="";
char *q=NULL,*ul=NULL,*tagstr=NULL,*sq=NULL,*mode=NULL;
char *sort=NULL;
char *env,*s,*t;
int tag=0,nwords=0,where=0,len=0,variables=0;
int i,j,lcindex=10000,sort_order=ORD_RATE;
int search_mode=MOD_ALL;

DOCUMENT * r=NULL, * r1=NULL;

	self=argv[0]?argv[0]:"search.cgi";
	tclock=time(0);srand(tclock);
	for(i=0;i<MAXRANDOM;i++)
		Randoms[i]=0;
	
	DbExitOnError=0;
	init_charset();

	printf("Content-type: text/html\r\n\r\n");

	/* Load Template */
	sprintf(str,"%s/%s", CONFIGDIR,(s=strrchr(self,'/'))?(s+1):(self));
	if(s=strrchr(str,'.')){
		*s=0;strcat(str,".htm");
	}else{
		strcpy(str,"search.htm");
	}
	file=fopen(str,"r");
	if(!file){
		printf("<html><body>Could not open template file '%s'!</body></html>\n",str);
		exit(0);
	}
	while((fgets(str,sizeof(str),file))&&(ntempl<MAXTEMPLATE)){
		if(!STRNCMP(str,"<!--top-->"))		where=T_TOP;
		else
		if(!STRNCMP(str,"<!--bottom-->"))	where=T_BOT;
		else
		if(!STRNCMP(str,"<!--restop-->"))	where=T_RESTOP;
		else
		if(!STRNCMP(str,"<!--resbot-->"))	where=T_RESBOT;
		else
		if(!STRNCMP(str,"<!--res-->"))		where=T_RESULT;
		else
		if(!STRNCMP(str,"<!--notfound-->"))	where=T_NOTFOUND;
		else
		if(!STRNCMP(str,"<!--error-->"))	where=T_ERROR;
		else
		if(!STRNCMP(str,"<!--clone-->"))	where=T_CLONE;
		else
		if(!STRNCMP(str,"<!--navigator-->"))	where=T_NAVIGATOR;
		else
		if(!STRNCMP(str,"<!--navleft-->"))	where=T_NAVLEFT;
		else
		if(!STRNCMP(str,"<!--navbar1-->"))	where=T_NAVBAR1;
		else
		if(!STRNCMP(str,"<!--navbar0-->"))	where=T_NAVBAR0;
		else
		if(!STRNCMP(str,"<!--navright-->"))	where=T_NAVRIGHT;
		else
		if(!STRNCMP(str,"<!--/top-->"))		where=0;
		else
		if(!STRNCMP(str,"<!--/bottom-->"))	where=0;
		else
		if(!STRNCMP(str,"<!--/restop-->"))	where=0;
		else
		if(!STRNCMP(str,"<!--/resbot-->"))	where=0;
		else
		if(!STRNCMP(str,"<!--/res-->"))		where=0;
		else
		if(!STRNCMP(str,"<!--/notfound-->"))	where=0;
		else
		if(!STRNCMP(str,"<!--/error-->"))	where=0;
		else
		if(!STRNCMP(str,"<!--/clone-->"))	where=0;
		else
		if(!STRNCMP(str,"<!--variables")){	where=0;variables=1;}
		else
		if(!STRNCMP(str,"<!--/navigator-->"))	where=0;
		else
		if(!STRNCMP(str,"<!--/navleft-->"))	where=0;
		else
		if(!STRNCMP(str,"<!--/navbar1-->"))	where=0;
		else
		if(!STRNCMP(str,"<!--/navbar0-->"))	where=0;
		else
		if(!STRNCMP(str,"<!--/navright-->"))	where=0;
		else
		if(!STRNCMP(str,"-->")){	
			if(variables){
				where=0;
				variables=0;
			}
		}else{
			if(where){
				Template[ntempl].where=where;
				Template[ntempl].str=strdup(str);
				ntempl++;
			}
			if((variables)&&(s=strtok(str,"\r\n"))){
				if(!STRNCASECMP(str,"MySQLHost="))
					DBHost=strdup(strchr(str,'=')+1);
				else
				if(!STRNCASECMP(str,"MySQLDB="))
					DBName=strdup(strchr(str,'=')+1);
				else
				if(!STRNCASECMP(str,"MySQLUser="))
					DBUser=strdup(strchr(str,'=')+1);
				else
				if(!STRNCASECMP(str,"MySQLPass="))
					DBPass=strdup(strchr(str,'=')+1);
				else
				if(!STRNCASECMP(str,"MySQLPort="))
					DBPort=atoi(strchr(str,'=')+1);
				else
				if(!STRNCASECMP(str,"DBHost="))
					DBHost=strdup(strchr(str,'=')+1);
				else
				if(!STRNCASECMP(str,"DBName="))
					DBName=strdup(strchr(str,'=')+1);
				else
				if(!STRNCASECMP(str,"DBUser="))
					DBUser=strdup(strchr(str,'=')+1);
				else
				if(!STRNCASECMP(str,"DBPass="))
					DBPass=strdup(strchr(str,'=')+1);
				else
				if(!STRNCASECMP(str,"DBPort="))
					DBPort=atoi(strchr(str,'=')+1);
				else
				if(!STRNCASECMP(str,"ResultsPerPage="))
					ps=atoi((strchr(str,'=')+1));
				else
				if(!STRNCASECMP(str,"PagesPerScreen="))
					pps=atoi((strchr(str,'=')+1));
				else
				if(!STRNCASECMP(str,"LocalCharset="))
					lcindex=getcharset((strchr(str,'=')+1));
				else
				if(!STRNCASECMP(str,"R")){
					i=atoi(str+1);
					s=strchr(str,'=');
					if(s){
						j=atoi(s+1);
						if((i>0)&&(i<MAXRANDOM)){
							frand=rand();
							frand=frand/RAND_MAX*j;
							Randoms[i]=frand;
						}
					}
				}
			}
		}
	}
	fclose(file);

	makewordchar(wordchar,lcindex);
	s=SEPARATOR;*s=0;
	for(i=1;i<255;i++){
		if(!strchr(wordchar,i)){
			*s=i;s++;*s=0;
		}
	}

	/* Parse Query String */
	env=getenv("QUERY_STRING");
	if(env){
		s=strtok(env,"&");
		while(s){
			if(!STRNCMP(s,"q=")){
				/*sq=strdup(s+2);*/
				q=strdup(Unescape(str,s+2));
				sq=strdup(EscapeURL(str,q));
				query_form_escaped=strdup(Escape(str,q,'"'));
			}else
			if(!STRNCMP(s,"np="))np=atoi(s+3);
			else
			if(!STRNCMP(s,"ps="))ps=atoi(s+3);
			else
			if((!STRNCMP(s,"ul="))&&(*(s+3))){
				sprintf(str1,"%%%s%%",Unescape(str,s+3));
				AddURLLimit(str1);ul=strdup(s+3);
			}else
			if((!STRNCMP(s,"t="))&&(*(s+2))){
				AddTagLimit(atoi(s+2));
			}else
			if(!STRNCMP(s,"m=")){
				mode=s+2;
				if(!strcmp(mode,"all"))search_mode=MOD_ALL;
				if(!strcmp(mode,"any"))search_mode=MOD_ANY;
			}else
			if(!STRNCMP(s,"s=")){
				sort=s+2;
				if(!strcmp(sort,"date"))sort_order=ORD_DATE;
				if(!strcmp(sort,"rate"))sort_order=ORD_RATE;
			}
			s=strtok(NULL,"&");
		}
	}

	PrintTemplate(T_TOP,"","","");
	if(!q){
		PrintTemplate(T_BOT,"","","");
		exit(0);
	}

	tolower_string(q,lcindex);
	if(r=Find(q,np,ps?ps:20,search_mode,sort_order,wordinfo,&found)){
		last=first=docnum=(ps?ps:20)*np+1;
		r1=r;while((++r1)->url_id)last++;
		num_pages = found/(ps?ps:20);
		is_next=last<found;
		if((ps?ps:20)*num_pages<found)num_pages++;
		PrintTemplate(T_RESTOP,"","","");
		sprintf(href,"%s?q=%s",self,sq);
		if(tagstr)sprintf(STREND(href),"&t=%s",tagstr);
		if(ul)sprintf(STREND(href),"&ul=%s",ul);
		if(mode)sprintf(STREND(href),"&m=%s",mode);
		if(sort)sprintf(STREND(href),"&s=%s",sort);
		if(ps)sprintf(STREND(href),"&ps=%d",ps);

		while(r->url_id){
			url_id=r->url_id;
			title=htmlspecialchars(r->title);
			text=htmlspecialchars(r->text);
			size=r->size;
			rating=r->rating;
			descr=htmlspecialchars(r->description);
			keywords=htmlspecialchars(r->keywords);
			crc=r->crc;
			PrintTemplate(T_RESULT,r->url,r->content_type,r->last_modified);
			free_it(title);free_it(text);
			free_it(descr);free_it(keywords);
			r++;docnum++;
		}
		PrintTemplate(T_RESBOT,"","","");
	}else{
		if(DbLastErrorCode){
			strcpy(error,DBErrorMsg());
			PrintTemplate(T_ERROR,"","","");
		}else	
			PrintTemplate(T_NOTFOUND,"","","");
	}
	PrintTemplate(T_BOT,"","","");
}
