#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/time.h>
#include "config.h"
#include "common.h"
#include "udm_syslog.h"
#include "dbms.h"
#include "charset.h"
#if (HAVE_MYSQL)
#include "mysql.h"
#elif (HAVE_PGSQL)
#include "libpq-fe.h"
#elif HAVE_MSQL
#include "msql.h"
#else
#endif

#define URL_DELETE_CACHE 50
#define URL_SELECT_CACHE 16

char *DBHost       = NULL;
char *DBName       = NULL;
char *DBUser       = NULL;
char *DBPass       = NULL;
int  DBPort        = 0;
int  DbLastErrorCode=0;
int  DbExitOnError=1;

static char DBErrorBuf[STRSIZ]="";
static char qbuf[STRSIZ]="";
static char get_url_buf[STRSIZ]="";
static int is_connected=0;
static int currow=1;
static int nrows=0;
static DOCUMENT doc;

/************************ MYSQL **************************************/
#if   (HAVE_MYSQL)
char DBVersion[]="MySQL";
static MYSQL mysql;
static MYSQL_RES *urlres=NULL;
static MYSQL_RES *res=NULL;
static MYSQL_RES *res1=NULL;
static MYSQL_ROW row;
#define SQL_NUM_ROWS(x)	mysql_num_rows(x)
#define SQL_FREE(x)	mysql_free_result(x)
#define lock_dict()	sql_query("LOCK TABLES dict WRITE")
#define flush_dict()	sql_query("UNLOCK TABLES");sql_query("LOCK TABLES dict WRITE")
#define unlock_dict()	sql_query("UNLOCK TABLES")
#define lock_url()	sql_query("LOCK TABLES url WRITE")
#define unlock_url()	sql_query("UNLOCK TABLES")

int connect_db()
{
	mysql_init(&mysql);
	if (!(mysql_real_connect(&mysql,DBHost,DBUser,DBPass,DBName,DBPort,NULL,0))){
		DbLastErrorCode=1;
		abort_indexer();
		return(1);
	}
	is_connected=1;return(0);
}
char * sql_value(MYSQL_RES *res,int i,int j){
MYSQL_ROW row;
	if(i>=mysql_num_rows(res))return(0);		
	mysql_data_seek(res,i);
	row=mysql_fetch_row(res);
	if(row)return(row[j]);
	else	return(0);
}
char * escstr(char *x,char *y){
	mysql_escape_string(x,y,strlen(y));
	return(x);
}
int exit_nicely(int c){
	if(is_connected)mysql_close(&mysql);
	is_connected=0;exit(c);
}
char *DBErrorMsg(){
	sprintf(DBErrorBuf,"Error #%d: '%s'",mysql_errno(&mysql),mysql_error(&mysql));
	return(DBErrorBuf);
}
static int safe_mysql_query(char *query)
{
int err,i;
	if(!is_connected)connect_db();
	if(DbLastErrorCode)return(DbLastErrorCode);
	for(i=0;i<2;i++){
		if(err=mysql_query(&mysql,query)){
			if((mysql_errno(&mysql)==CR_SERVER_LOST)||
			(mysql_errno(&mysql)==CR_SERVER_GONE_ERROR)||
			(mysql_errno(&mysql)==ER_SERVER_SHUTDOWN)){
				udm_log(UDM_LOG_ERROR,"Error in mysql_query: %d (%s)",mysql_errno(&mysql),mysql_error(&mysql));
				udm_log(UDM_LOG_ERROR,"Waiting during 5 seconds for server to restart...");
				sleep(5);
				udm_log(UDM_LOG_ERROR,"Trying to reconnect");
			}else
				return(err);
		}else
			return(err);
	}
	return(err);
}
MYSQL_RES * sql_query(char *query)
{
	DbLastErrorCode=0;
	if(safe_mysql_query(query)){
		DbLastErrorCode=1;
		if(mysql_errno(&mysql)!=ER_DUP_ENTRY)abort_indexer();
		return(0);
	}
	return(mysql_store_result(&mysql));
}

/*********************************** POSTGRESQL *********************/
#elif (HAVE_PGSQL)
char DBVersion[]="PostgreSQL";
static PGconn     *pgsql=NULL;
static PGresult   *res=NULL;
static PGresult   *res1=NULL;
static PGresult   *urlres=NULL;
#define SQL_NUM_ROWS(x)	PQntuples(x)
#define SQL_FREE(x)	PQclear(x)
#define sql_value(x,y,z) PQgetvalue(x,y,z)
#define lock_dict()	sql_query("BEGIN WORK")
#define	flush_dict()	sql_query("END WORK");sql_query("BEGIN WORK")
#define unlock_dict()	sql_query("END WORK")
#define lock_url()	sql_query("BEGIN WORK");sql_query("LOCK url")
#define unlock_url()	sql_query("END WORK")
char * escstr(char *x,char *y){
char *s;
	s=x;
	while(*y){
		switch(*y){
			case '\'':*x='\\';x++;
			default:*x=*y;
		}
		x++;y++;
	}
	*x=0;return(s);
}
int connect_db()
{
char port[8];
	sprintf(port,"%d",DBPort);
	pgsql = PQsetdbLogin(DBHost, DBPort?port:0, 0, 0, DBName, DBUser, DBPass);
	if (PQstatus(pgsql) == CONNECTION_BAD){
		DbLastErrorCode=1;
		abort_indexer();
		return(1);
	}
	is_connected=1;return(0);
}
int exit_nicely(int c){
	if(is_connected)PQfinish(pgsql);
	is_connected=0;exit(c);
}
char *DBErrorMsg(){
	sprintf(DBErrorBuf, "Error: '%s'", PQerrorMessage(pgsql));
	return(DBErrorBuf);
}
PGresult * safe_pgsql_query(char *q){
	if(!is_connected){
		connect_db();
		if(DbLastErrorCode)return(0);
		res=PQexec(pgsql,"SET TIME ZONE 'GMT'");
		if(PQresultStatus(res)!= PGRES_COMMAND_OK){
			abort_indexer();
			DbLastErrorCode=1;
			return(0);
		}
	}
	return(PQexec(pgsql,q));
}
PGresult * sql_query(char *q){
PGresult * res;
	
	DbLastErrorCode=0;res=safe_pgsql_query(q);
	if((PQresultStatus(res)!=PGRES_COMMAND_OK)&&(PQresultStatus(res)!=PGRES_TUPLES_OK)){
		DbLastErrorCode=1;
		if(!strstr(PQerrorMessage(pgsql),"duplicate"))abort_indexer();
		return(0);
	}
	return(res);
}

/*****************************  miniSQL *******************************/
#elif (HAVE_MSQL)
char DBVersion[]="miniSQL";
int			msql=0;
static m_result		*res=NULL;
static m_result		*res1=NULL;
static m_result		*urlres=NULL;
#define SQL_NUM_ROWS(x)	msqlNumRows(x)
#define SQL_FREE(x)	msqlFreeResult(x)
#define lock_dict()	
#define	flush_dict()	
#define unlock_dict()	
#define lock_url()	
#define unlock_url()
char * sql_value(m_result *res,int i,int j){
m_row row;
	if(i>=msqlNumRows(res))return("");
	msqlDataSeek(res,i);
	row=msqlFetchRow(res);
	if(row)return(row[j]?row[j]:"");
	else	return(0);
}
char * escstr(char *x,char *y){
char *s;
	s=x;
	while(*y){
		switch(*y){
			case '\'':*x='\\';x++;
			default:*x=*y;
		}
		x++;y++;
	}
	*x=0;return(s);
}
int connect_db(){
char *host;
	if(host=DBHost)if(!strcmp(DBHost,"localhost"))host=NULL;
	if((msql=msqlConnect(host))<0){
		DbLastErrorCode=1;
		abort_indexer();
		return(1);
	}
	if(msqlSelectDB(msql,DBName)==-1){
		DbLastErrorCode=1;
		abort_indexer();
		return(1);
	}
	is_connected=1;return(0);
}
int exit_nicely(int c){
	if(is_connected)msqlClose(msql);
	is_connected=0;exit(c);
}
char *DBErrorMsg(){
	sprintf(DBErrorBuf,"Error: '%s'",msqlErrMsg);
	return(DBErrorBuf);
}
m_result * sql_query(char *query);
static int safe_msql_query(char *query)
{
int err,i;
	if(!is_connected){
		connect_db();
		if(DbLastErrorCode)
			return(DbLastErrorCode);
	}
	/*printf("'%s'\n",query);*/
	if((err=msqlQuery(msql,query))==-1){
		if(!strstr(msqlErrMsg,"Non unique")){
			abort_indexer();
			DbLastErrorCode=1;
			return(err);
		}else{
			/*printf("Not Uniq!\n");*/
		}
		return(0);
	}
	return(0);
}
m_result * sql_query(char *query)
{
int err;
	DbLastErrorCode=0;
	if(err=safe_msql_query(query)){
		DbLastErrorCode=1;
		return(0);
	}
	return(msqlStoreResult());
}

/******************************** Template  ************************/
#else
char DBVersion[]="(not selected)";
static int   *res=NULL;
static int   *res1=NULL;
static int   *urlres=NULL;
#define SQL_NUM_ROWS(x)	0
#define SQL_FREE(x)
#define sql_value(x,y,z) 0
#define lock_dict()
#define	flush_dict()
#define unlock_dict()
#define lock_url()
#define unlock_url()
char * escstr(char *x,char *y){
	strcpy(x,y);return(x);
}
int connect_db(){
	is_connected=1;return(0);
}
int exit_nicely(int c){
	exit(c);
}
char *DBErrorMsg(){
	sprintf(DBErrorBuf, "Error");
	return(DBErrorBuf);
}
static int safe_msql_query(char *query){
	return(0);
}
int * sql_query(char *query){
	return(0);
}
#endif

/********************************************************************/
/* --------------------------------------------------------------- */
/* Almost Unified code for miniSQL,PostgreSQL and MySQL            */
/* --------------------------------------------------------------- */
/*******************************************************************/

time_t now(){
time_t tloc;
	return(time(&tloc));
}

int abort_indexer()
{
	if(DbExitOnError){
		if(is_log_open)udm_log(UDM_LOG_ERROR,"%s",DBErrorMsg());
		else	fprintf(stderr,"%s\n",DBErrorMsg());
		exit_nicely(1);
	}
}

static char tagstr[STRSIZ]="";
static char statusstr[BUFSIZ]="";
static char urlstr[BUFSIZ]="";

int AddTagLimit(int tag){
#if (HAVE_MYSQL || HAVE_PGSQL)
	if(*tagstr)sprintf(STREND(tagstr)-1,",%d)",tag);
	else	sprintf(tagstr," AND url.tag IN (%d)",tag);
#elif (HAVE_MSQL)
	if(*tagstr)strcpy(STREND(tagstr)-1," OR ");
	else	strcat(tagstr," AND (");
	sprintf(STREND(tagstr),"url.tag=%d)",tag);
#endif
	return(0);
}

int AddStatusLimit(int status){
#if (HAVE_MYSQL || HAVE_PGSQL)
	if(*statusstr)sprintf(STREND(statusstr)-1,",%d)",status);
	else	sprintf(statusstr," AND url.status IN (%d)",status);
#elif (HAVE_MSQL)
	if(*statusstr)strcpy(STREND(statusstr)-1," OR ");
	else	strcat(statusstr," AND (");
	sprintf(STREND(statusstr),"url.status=%d)",status);
#endif
	return(0);
}

int AddURLLimit(char *URL){
	if(*urlstr)strcpy(STREND(urlstr)-1," OR ");
	else	strcat(urlstr," AND (");
	sprintf(STREND(urlstr),"url.url LIKE '%s')",URL);
}


char ** LoadStopList(){
int i=0;
int in_stop_list=0;
char **stoplist=NULL;

	res=sql_query("SELECT word FROM stopword");
	in_stop_list=SQL_NUM_ROWS(res);
	stoplist=(char **)malloc((in_stop_list+1)*sizeof(char *));
	for(i=0;i<in_stop_list;i++){
		stoplist[i]=strdup(sql_value(res,i,0));
	}
	SQL_FREE(res);
	stoplist[i]=0;
	return(stoplist);
}
int AddURL(char *url,int referrer,int hops){
char *e_url;
int next_url_id=0;
	e_url=(char*)malloc(strlen(url)*2);escstr(e_url,url);
#if (HAVE_MSQL)
	if(res=sql_query("SELECT _seq FROM url"))next_url_id=atoi(sql_value(res,0,0));
	else{
		abort_indexer();
		DbLastErrorCode=1;
		return(0);
	}
	sprintf(qbuf,"INSERT INTO url (url,referrer,hops,rec_id,crc,last_index_time,next_index_time) VALUES ('%s',%d,%d,%d,'',%d,%d)",e_url,referrer,hops,next_url_id,now(),now());
#else
	sprintf(qbuf,"INSERT INTO url (url,referrer,hops) VALUES ('%s',%d,%d)",e_url,referrer,hops);
#endif
	free(e_url);sql_query(qbuf);return(0);
}
int StoreHrefs(int hcur,HREF *Href){
int i;
	for(i=0;i<hcur;i++){
		if(!Href[i].stored)
			AddURL(Href[i].href,Href[i].referrer,Href[i].hops);
		Href[i].stored=1;
	}
}
int DeleteWordFromURL(int url_id){
	sprintf(qbuf,"DELETE FROM dict WHERE url_id=%d",url_id);
	sql_query(qbuf);return(0);
}
int DeleteUrl(int url_id){
	sprintf(qbuf,"DELETE FROM dict WHERE url_id=%d",url_id);
	sql_query(qbuf);
	sprintf(qbuf,"DELETE FROM url WHERE rec_id=%d",url_id);
	sql_query(qbuf);return(0);
}

ROBOT * LoadRobots(){
ROBOT *Robots=0;
int i,nrobots;
	res=sql_query("SELECT hostinfo,path FROM robots");
	nrobots=SQL_NUM_ROWS(res);
	if(nrobots){
		Robots=(ROBOT *)malloc((nrobots+1)*sizeof(ROBOT));
		for(i=0;i<nrobots;i++){
			Robots[i].hostinfo=strdup(sql_value(res,i,0));
			Robots[i].path=strdup(sql_value(res,i,1));
		}
		Robots[i].hostinfo=Robots[i].path=NULL;
	}
	SQL_FREE(res);
	return(Robots);
}
int DeleteRobotsFromHost(char *hostinfo){
	sprintf(qbuf,"DELETE FROM robots WHERE hostinfo='%s'",hostinfo);
	sql_query(qbuf);
	return(0);
}
int AddRobotsToHost(char *hostinfo,char *s){
	sprintf(qbuf,"INSERT INTO robots (hostinfo,path) VALUES ('%s','%s')",hostinfo,s);
	sql_query(qbuf);
	return(0);
}
int FindOrigin(char *crc){
int i=0,origin=0;
char *o;
	if(!crc)return(0);if(!(*crc))return(0);
#if (HAVE_MYSQL || HAVE_PGSQL )
	sprintf(qbuf,"SELECT min(rec_id) FROM url WHERE crc='%s'",crc);
	res=sql_query(qbuf);
	if(SQL_NUM_ROWS(res)==1)
		if(o=sql_value(res,0,0))
			origin=atoi(o);
	SQL_FREE(res);
#elif (HAVE_MSQL)
	sprintf(qbuf,"SELECT rec_id FROM url WHERE crc='%s'",crc);
	res=sql_query(qbuf);
	for(i=0;i<SQL_NUM_ROWS(res);i++)
		if(o=sql_value(res,i,0))
			if((!origin)||(origin>atoi(o)))
				i=atoi(o);
	SQL_FREE(res);
#endif
	return(origin);
}

int StoreWords(int url_id, int wcur,WORD *Word)
{
WORD * findword(int wcur,WORD *Word,char *w){
int i;
	for(i=0;i<wcur;i++)
		if(!strcmp(Word[i].word,w))
			return(&Word[i]);
	return(0);
}
int i,old,new,flush;
int were,added,deleted,updated;
char *s;WORD *w;
	flush=were=added=deleted=updated=0;
	sprintf(qbuf,"SELECT word,intag FROM dict WHERE url_id=%d",url_id);
	res=sql_query(qbuf);
	lock_dict();
	were=SQL_NUM_ROWS(res);
	for(i=0;i<were;i++){
		s=sql_value(res,i,0);
		old=atoi(sql_value(res,i,1));
		if(w=findword(wcur,Word,s)){
			new=w->count;
			if((new)&&(old)&&(new!=old)){
				sprintf(qbuf,"UPDATE dict SET intag=%d WHERE word='%s' AND url_id='%d'",new,Word[i].word,url_id);
				sql_query(qbuf);
				Word[i].count=0;updated++;
				flush++;
			}
			w->count=0;
		}else{
			sprintf(qbuf,"DELETE FROM dict WHERE url_id=%d AND word='%s'",url_id,s);
			sql_query(qbuf);
			deleted++;flush++;
		}
		if(flush>1024){
			flush_dict();
			flush=0;
		}
	}
	SQL_FREE(res);
	for(i=0;i<wcur;i++){
		if(Word[i].count){
			sprintf(qbuf,"INSERT INTO dict (url_id,word,intag) VALUES(%d,'%s',%d)",url_id,Word[i].word,Word[i].count);
			sql_query(qbuf);
			flush++;added++;
			if(flush>1024){
				flush_dict();
				flush=0;
			}
		}
	}
	unlock_dict();
	return(deleted||added||updated);
}
int DeleteAllFromDict(){
	sprintf(qbuf,"DELETE FROM dict");
	sql_query(qbuf);return(0);
}
int DeleteAllFromUrl(){
	sprintf(qbuf,"DELETE FROM url");
	sql_query(qbuf);return(0);
}
int MarkForReindex(){
#if (HAVE_MYSQL || HAVE_PGSQL)
	sprintf(qbuf,"UPDATE url SET next_index_time=NOW() WHERE 1=1 %s%s%s",tagstr,urlstr,statusstr);
	sql_query(qbuf);return(0);
#elif (HAVE_MSQL)
	sprintf(qbuf,"UPDATE url SET next_index_time=%d WHERE rec_id>=0 %s%s%s",now(),tagstr,urlstr,statusstr);
	sql_query(qbuf);return(0);
#endif
	
}



int  MarkClone(int url_id,int status,int period,
char *content_type,char *last_modified_escaped,char *digest){
#if (HAVE_MYSQL)
	sprintf(qbuf,
"UPDATE url \
SET crc='%s',status=%d,\
next_index_time=from_unixtime(unix_timestamp()+%d), \
content_type='%s',last_modified='%s' \
WHERE rec_id=%d",digest,status,period,content_type,last_modified_escaped,url_id);
	sql_query(qbuf);
#elif (HAVE_PGSQL)
	sprintf(qbuf,
"UPDATE url \
SET crc='%s',status=%d,\
next_index_time=(now()::datetime+'%d secs'::timespan), \
content_type='%s',last_modified='%s' \
WHERE rec_id=%d",digest,status,period,content_type,last_modified_escaped,url_id);
	sql_query(qbuf);
#elif (HAVE_MSQL)
	sprintf(qbuf,
"UPDATE url \
SET crc='%s',status=%d,content_type='%s',last_modified='%s',next_index_time=%d \
WHERE rec_id=%d",digest,status,content_type,last_modified_escaped,now()+period,url_id);
	sql_query(qbuf);
#endif
	return(0);
}

int UpdateUrl(int url_id,int status,int period){
#if (HAVE_MYSQL)
	sprintf(qbuf,"UPDATE url SET status=%d,next_index_time=from_unixtime(unix_timestamp()+%d) WHERE rec_id='%d'",status,period,url_id);
	sql_query(qbuf);
#elif (HAVE_PGSQL)
	sprintf(qbuf,"UPDATE url SET status=%d,next_index_time=(now()::datetime+'%d secs'::timespan) WHERE rec_id='%d'",status,period,url_id);
	sql_query(qbuf);
#elif (HAVE_MSQL)
	sprintf(qbuf,"UPDATE url SET status=%d,next_index_time=%d WHERE rec_id=%d",status,now()+period,url_id);
	sql_query(qbuf);
#endif
	
}

int LongUpdateUrl(
int url_id,int status,int changed,int size,int period,
int hint,int index,
char *last_modified_escaped,
char *text_escaped,
char *title_escaped,
char *content_type,
char *keywords_escaped,
char *descript_escaped,
char *digest){

#if (HAVE_MYSQL)
	sprintf(qbuf,"\
UPDATE url SET \
status=%d,last_modified='%s',\
next_index_time=from_unixtime(unix_timestamp()+%d),\
tag=%d,text='%s',title='%s',content_type='%s',size=%d,\
indexed=%d,keywords='%s',description='%s',crc='%s'%s \
WHERE rec_id=%d",
	status,last_modified_escaped,
	period,hint,text_escaped,title_escaped,content_type,size,index,
	keywords_escaped,descript_escaped,digest,
	changed?",last_index_time=NOW()":"",url_id);
	sql_query(qbuf);
#elif (HAVE_PGSQL)
sprintf(qbuf,"\
UPDATE url SET \
status=%d,last_modified='%s',\
next_index_time=(now()::datetime+'%d secs'::timespan),\
tag=%d,text='%s',title='%s',content_type='%s',size=%d,\
indexed=%d,keywords='%s',description='%s',crc='%s'%s \
WHERE rec_id=%d",
	status,last_modified_escaped,period,
	hint,text_escaped,title_escaped,content_type,size,index,
	keywords_escaped,descript_escaped,digest,
	changed?",last_index_time=NOW()":"",url_id);
	sql_query(qbuf);
#elif (HAVE_MSQL)
char last_index_time[64]="";

	if(changed)sprintf(last_index_time,",last_index_time=%d",now());

sprintf(qbuf,"\
UPDATE url SET \
status=%d,last_modified='%s',\
next_index_time=%d,\
tag=%d,txt='%s',title='%s',content_type='%s',size=%d,\
keywords='%s',description='%s',crc='%s',\
next_index_time=%d%s \
WHERE rec_id=%d",
	status,last_modified_escaped,now()+period,
	hint,text_escaped,title_escaped,content_type,size,
	keywords_escaped,descript_escaped,digest,now()+period,
	last_index_time,url_id);
	sql_query(qbuf);
#endif
	return(0);
}


DOCUMENT *GetDocInfo(int gmt_offset,int expired_first){

char urlin[STRSIZ]="";
char sortstr[64]="";
int i=0;
	sprintf(sortstr,expired_first?" ORDER BY next_index_time":"");
	if(currow>=nrows){
		if(urlres)SQL_FREE(urlres);
		lock_url();
#if (HAVE_MYSQL)
		sprintf(get_url_buf,
		"SELECT url,rec_id,size,status,concat(date_format(from_unixtime(unix_timestamp(last_index_time)+(%d)),'%%a, %%e %%b %%Y %%T'),' GMT'),hops,crc FROM url WHERE next_index_time<=NOW() %s%s%s%s LIMIT %d",
			gmt_offset,tagstr,urlstr,statusstr,sortstr,URL_SELECT_CACHE);
#elif (HAVE_PGSQL)
		sprintf(get_url_buf,
		"SELECT url,rec_id,size,status,last_index_time,hops,crc FROM url WHERE (next_index_time<=NOW()::datetime) %s%s%s%s LIMIT %d",
			tagstr,urlstr,statusstr,sortstr,URL_SELECT_CACHE);
#elif (HAVE_MSQL)
		sprintf(get_url_buf,
		"SELECT url,rec_id,size,status,last_index_time,hops,crc,last_modified FROM url WHERE next_index_time<=%d %s%s%s%s LIMIT %d",
			now(),tagstr,urlstr,statusstr,sortstr,URL_SELECT_CACHE);
#endif
		urlres=sql_query(get_url_buf);
		nrows=SQL_NUM_ROWS(urlres);currow=0;
		if(!nrows){
			unlock_url();
			/*if(urlres)SQL_FREE(urlres);*/
			return(0);
		}
		urlin[0]=0;
		for(i=0;i<nrows;i++){
#if (HAVE_MSQL)
			sprintf(qbuf,"UPDATE url SET next_index_time=%d WHERE rec_id=%d",now()+3600,atoi(sql_value(urlres,i,1)));
			sql_query(qbuf);
#elif (HAVE_MYSQL || HAVE_PGSQL)
			if(urlin[0])strcat(urlin,",");
			strcat(urlin,sql_value(urlres,i,1));
#endif
		}
#if (HAVE_MYSQL)
		sprintf(qbuf,"UPDATE url SET next_index_time=from_unixtime(unix_timestamp()+3600) WHERE rec_id in (%s)",urlin);
		sql_query(qbuf);
#elif (HAVE_PGSQL)
		sprintf(qbuf,"UPDATE url SET next_index_time=(now()::datetime+'3600 secs'::timespan) WHERE rec_id in (%s)",urlin);
		sql_query(qbuf);
#endif
		unlock_url();
	}
	doc.url=sql_value(urlres,currow,0);
	doc.url_id=atoi(sql_value(urlres,currow,1));
	doc.size=atoi(sql_value(urlres,currow,2));
	doc.status=atoi(sql_value(urlres,currow,3));
	doc.last_index_time=sql_value(urlres,currow,4);
	doc.hops=atoi(sql_value(urlres,currow,5));
	doc.crc=sql_value(urlres,currow,6)?sql_value(urlres,currow,6):"___";
	doc.last_modified=sql_value(urlres,currow,7);
	currow++;
	return(&doc);
}



/************************************************************/
/* Misc functions                                           */
/************************************************************/

int AddSpell(char *flag,char *lang,char *word){
	sprintf(qbuf,"INSERT INTO spell (word,flag,lang) VALUES ('%s','%s','%s')",word,flag,lang);
	sql_query(qbuf);return(DbLastErrorCode);
}

int AddAffix(char *flag,char *lang,char *mask,char *find,char *repl,char *type){
	sprintf(qbuf,"INSERT INTO affix (flag,type,lang,mask,find,repl) VALUES ('%s','%s','%s','%s$','%s','%s')",flag,type,lang,mask,find,repl);
	sql_query(qbuf);return(DbLastErrorCode);
}

STAT * GetStatistics(){
int i,j,n;
STAT *stat=NULL;

#if (HAVE_MYSQL||HAVE_PGSQL)
#if   (HAVE_MYSQL)
	sprintf(qbuf,"SELECT status,sum(next_index_time<=NOW()),count(*) FROM url WHERE 1=1 %s%s%s GROUP BY status",tagstr,urlstr,statusstr);
#elif (HAVE_PGSQL)
	sprintf(qbuf,"SELECT status,sum(case when next_index_time<=NOW() then 1 else 0 end),count(*) FROM url WHERE 1=1 %s%s%s GROUP BY status",tagstr,urlstr,statusstr);
#endif
	res=sql_query(qbuf);
	if(j=SQL_NUM_ROWS(res)){
		stat=(STAT *)malloc(sizeof(STAT)*(j+1));
		for(i=0;i<j;i++){
			stat[i].status=atoi(sql_value(res,i,0));
			stat[i].expired=atoi(sql_value(res,i,1));
			stat[i].total=atoi(sql_value(res,i,2));
		}
		stat[i].status=-1;
	}
	SQL_FREE(res);
#elif (HAVE_MSQL)
	n=now();
	sprintf(qbuf,"SELECT status,next_index_time FROM url WHERE rec_id>=0 %s%s%s",tagstr,urlstr,statusstr);
	res=sql_query(qbuf);
	if(SQL_NUM_ROWS(res)){
		stat=(STAT *)malloc(sizeof(STAT)*(SQL_NUM_ROWS(res)+1));
		stat[0].status=-1;
		for(i=0;i<SQL_NUM_ROWS(res);i++){
			for(j=0;stat[j].status!=-1;j++){
				if(stat[j].status==atoi(sql_value(res,i,0))){
					if(atoi(sql_value(res,i,1))<=n)
						stat[j].expired++;
					stat[j].total++;
					break;
				}
			}
			if(stat[j].status==-1){
				stat[j].status=atoi(sql_value(res,i,0));
				stat[j].expired=0;
				if(atoi(sql_value(res,i,1))<=n)
					stat[j].expired++;
				stat[j].total=1;
				stat[j+1].status=-1;
			}
		}
	}
	SQL_FREE(res);
#endif
	return(stat);
}

int ShowReferers(){
int i,j;
	printf("\n          URLs and referers \n\n");
	sprintf(qbuf,"SELECT url2.url,url.url,url.status FROM url,url url2 WHERE url.referrer=url2.rec_id %s%s%s ",tagstr,urlstr,statusstr);
	res=sql_query(qbuf);
	j=SQL_NUM_ROWS(res);
	for(i=0;i<j;i++)
		printf("%s -> %s (%s)\n",sql_value(res,i,0),
			sql_value(res,i,1),sql_value(res,i,2));
	SQL_FREE(res);return(0);
}

int ClearDatabase(){
char urlin[STRSIZ]="";char str[STRSIZ]="";
int i,j;

	printf("You are going to delete database '%s' contents\n",DBName);
	printf("Are you sure?(YES/no)");
	if(fgets(str,sizeof(str),stdin)){
		if(strncmp(str,"YES",3)){
			printf("Canceled\n");
			return(0);
		}
	}else{
		printf("Canceled\n");
		return(0);
	}
	if(!tagstr[0]&&!urlstr[0]&&!statusstr[0]){
		printf("Deleting all records\n");
		DeleteAllFromDict();DeleteAllFromUrl();
		printf("Done\n");
	}else{
		printf("Deleting\n");j=0;
		while(1){
			sprintf(qbuf,"SELECT rec_id FROM url WHERE rec_id>=0 %s%s%s LIMIT %d",tagstr,urlstr,statusstr,URL_DELETE_CACHE);
			res=sql_query(qbuf);
			if(SQL_NUM_ROWS(res)){
#if (HAVE_MYSQL || HAVE_PGSQL)
				urlin[0]=0;
				for(i=0;i<SQL_NUM_ROWS(res);i++){
					if(i)strcat(urlin,",");
					strcat(urlin,sql_value(res,i,0));
				}
				j+=i;
				sprintf(qbuf,"DELETE FROM dict WHERE url_id in (%s)",urlin);
				sql_query(qbuf);
				sprintf(qbuf,"DELETE FROM url WHERE rec_id in (%s)",urlin);
				sql_query(qbuf);
				SQL_FREE(res);
#elif (HAVE_MSQL)
				for(i=0;i<SQL_NUM_ROWS(res);i++)
					DeleteUrl(atoi(sql_value(res,i,0)));
				j+=i;
#endif
				printf("%d records\r",j);fflush(stdout);
			}else{
				SQL_FREE(res);
				break;
			}
		}
		printf("\nDone\n");
	}
	return(0);
}

/******************* Search stuff ************************************/

DOCUMENT * CloneList(char *crc){
int i;
DOCUMENT * r=NULL;

	sprintf(qbuf,"SELECT rec_id,url,content_type,last_modified FROM url WHERE crc='%s'",crc);
	res=sql_query(qbuf);if(DbLastErrorCode)return(NULL);
	r=(DOCUMENT*)malloc(sizeof(DOCUMENT)*(SQL_NUM_ROWS(res)+1));
	for(i=0;i<SQL_NUM_ROWS(res);i++){
		r[i].url_id=atoi(sql_value(res,i,0));
		r[i].url=strdup(sql_value(res,i,1));
		r[i].content_type=strdup(sql_value(res,i,2));
		r[i].last_modified=strdup(sql_value(res,i,3));
		r[i+1].url_id=0;
	}
	SQL_FREE(res);
	return(r);
}


char ** GetAllWordForms(char *word){
int i,j,l,n1,n2;
char **list=NULL;

int AddInList(char *w){
char **i;
	i=list;
	while(*i){
		if(!strcmp(*i,w))return(0);
		i++;
	}
	*i=strdup(w);i++;*i=0;return(0);
}
	list=(char **)malloc((sizeof(char **))*128);*list=NULL;
	l=strlen(word);AddInList(word);

#ifdef HAVE_PGSQL
	sprintf(qbuf,
"select trim((substr('%s',1,%d-length(a.repl))||a.find)) as res \
from affix as a,spell as s where a.type='s' and position(a.flag in s.flag)>0 and \
trim((substr('%s',1,%d-length(a.repl))||a.find))=s.word and \
trim(substr('%s',%d+1-length(a.repl),length(a.repl)))=trim(a.repl) and \
trim((substr('%s',1,%d-length(a.repl))||a.find)) ~* trim(a.mask) \
and a.lang=s.lang",
	word,l,word,l,word,l,word,l);

	res=sql_query(qbuf);if(DbLastErrorCode)return(NULL);
	n1=SQL_NUM_ROWS(res);
	for(i=-1;i<n1;i++){
		if(i>-1)AddInList(sql_value(res,i,0));
		sprintf(qbuf,
"select distinct \
trim(substr(spell.word,1,length(spell.word)-length(affix.find))||repl) \
from spell,affix where spell.word='%s' \
and position(affix.flag IN spell.flag)>0 \
and spell.lang=affix.lang and affix.type='s' \
and trim(spell.word) ~* trim(affix.mask)",
		i==-1?word:sql_value(res,i,0));
		res1=sql_query(qbuf);if(DbLastErrorCode)return(NULL);
		n2=SQL_NUM_ROWS(res1);
		for(j=0;j<n2;j++)AddInList(sql_value(res1,j,0));
	}
	SQL_FREE(res1);
	SQL_FREE(res);
#else
#ifdef HAVE_MYSQL
	sprintf(qbuf,
"select s.word \
from affix as a,spell as s \
where right('%s',length(a.repl))=a.repl \
and a.type='s' \
and a.lang=s.lang \
and INSTR(s.flag,a.flag) \
and concat(left('%s',%d-length(a.repl)),a.find)=s.word \
and s.word regexp a.mask",
	word,word,l);

	res=sql_query(qbuf);if(DbLastErrorCode)return(NULL);
	n1=SQL_NUM_ROWS(res);
	for(i=-1;i<n1;i++){
		if(i>-1)AddInList(sql_value(res,i,0));
		sprintf(qbuf,
"select distinct concat(left(word,length(word)-length(find)),repl) as res \
from spell,affix \
where spell.word='%s' \
and INSTR(spell.flag,affix.flag) and \
spell.lang=affix.lang \
and affix.type='s' \
and word regexp mask",
		i==-1?word:sql_value(res,i,0));
		res1=sql_query(qbuf);if(DbLastErrorCode)return(NULL);
		n2=SQL_NUM_ROWS(res1);
		for(j=0;j<n2;j++)AddInList(sql_value(res1,j,0));
	}
	SQL_FREE(res1);
	SQL_FREE(res);
#endif
#endif
	return(list);
}


#if (HAVE_MYSQL || HAVE_PGSQL)
DOCUMENT * Find(char *q,int np,int ps,int search_mode,int sort_order,char *wordinfo,int *found){
char **list,**o,*w;
char wordin[STRSIZ]="";
char curwordin[STRSIZ]="";
char urlin[STRSIZ]="";
char countstr[STRSIZ]="";
char orderstr[STRSIZ]="";
char havingstr[STRSIZ]="";
int i,j,n,first,last,word_number=0;
DOCUMENT * r=NULL;

	w=strtok(q,SEPARATOR);
	while(w){
		word_number++;
		o=list=GetAllWordForms(w);
		if(DbLastErrorCode)return(NULL);
		if(!list)return(0);
		curwordin[0]=0;
		while(*list){
			if(*wordin)strcat(wordin,",");
			sprintf(STREND(wordin),"'%s'",*list);
			if(*curwordin)strcat(curwordin,",");
			sprintf(STREND(curwordin),"'%s'",*list);
			free_it(*list);list++;
		}
		free_it(o);
		if(*countstr)strcat(countstr,"+");
#if (HAVE_MYSQL)
		sprintf(STREND(countstr),"max(if(word IN (%s),1,0))",curwordin);
#elif (HAVE_PGSQL)
		sprintf(STREND(countstr),"max(case when word IN (%s) then 1 else 0 end)",curwordin);
#endif
		w=strtok(NULL,SEPARATOR);
	}

	sprintf(qbuf,"SELECT dict.word,count(*)as r FROM dict,url WHERE url.rec_id=dict.url_id AND dict.word IN (%s) %s%s%s GROUP BY dict.word ORDER BY r DESC",wordin,tagstr,urlstr,statusstr);
	res=sql_query(qbuf);if(DbLastErrorCode)return(NULL);
	for(i=0;i<SQL_NUM_ROWS(res);i++){
		strcat(wordinfo," ");strcat(wordinfo,sql_value(res,i,0));
		strcat(wordinfo,":");strcat(wordinfo,sql_value(res,i,1));
	}
	SQL_FREE(res);

	if(search_mode==MOD_ALL)sprintf(havingstr," HAVING num=%d",word_number);
	if(sort_order==ORD_RATE)strcpy(orderstr," ORDER BY num DESC, rate DESC");
	if(sort_order==ORD_DATE)strcpy(orderstr," ORDER BY last_index_time DESC");
	sprintf(qbuf,"SELECT dict.url_id,%s as num,sum(intag)as rate FROM dict,url WHERE url.rec_id=dict.url_id AND dict.word IN (%s) %s%s%s GROUP BY dict.url_id %s%s",countstr,wordin,urlstr,statusstr,tagstr,havingstr,orderstr);
	res=sql_query(qbuf);if(DbLastErrorCode)return(NULL);
	*found=n=SQL_NUM_ROWS(res);
	if(!n){
		SQL_FREE(res);
		return(0);
	}
	first=np*ps;last=first+ps;
	if(last>n)last=n;n=last-first;

	r=(DOCUMENT*)malloc(sizeof(DOCUMENT)*(n+1));
	r[0].url_id=0;
	for(i=0;i<n;i++){
		r[i+1].url_id=0;
		r[i].url_id=atoi(sql_value(res,i+first,0));
		r[i].rating=atoi(sql_value(res,i+first,1));
		if(*urlin)strcat(urlin,",");
		strcat(urlin,sql_value(res,i+first,0));
		r[i].url=r[i].content_type=r[i].last_modified=0;
		r[i].title=r[i].text=r[i].last_index_time=0;
		r[i].keywords=r[i].description=r[i].crc=0;
		r[i].size=r[i].referrer=r[i].hops;
	}
	SQL_FREE(res);
	sprintf(qbuf,"SELECT \
rec_id,url,content_type,last_modified,title,text, \
size,last_index_time,next_index_time, \
referrer,keywords,description,crc FROM url \
WHERE rec_id IN (%s)",urlin);

	res=sql_query(qbuf);if(DbLastErrorCode)return(NULL);
	for(i=0;i<n;i++)
	for(j=0;j<SQL_NUM_ROWS(res);j++)
	if(r[i].url_id==atoi(sql_value(res,j,0))){
		r[i].url=strdup(sql_value(res,j,1));
		r[i].content_type=strdup(sql_value(res,j,2));
		r[i].last_modified=strdup(sql_value(res,j,3));
		r[i].title=strdup(sql_value(res,j,4));
		r[i].text=strdup(sql_value(res,j,5));
		r[i].size=atoi(sql_value(res,j,6));
		r[i].last_index_time=strdup(sql_value(res,j,7));
		r[i].next_index_time=strdup(sql_value(res,j,8));
		r[i].referrer=atoi(sql_value(res,j,9));
		r[i].keywords=strdup(sql_value(res,j,10));
		r[i].description=strdup(sql_value(res,j,11));
		r[i].crc=strdup(sql_value(res,j,12));
	}
	SQL_FREE(res);
	return(r);
}
#elif (HAVE_MSQL)
DOCUMENT * Find(char *q,int np,int ps,int search_mode,int sort_order,char *wordinfo,int *found){
DOCUMENT * r=NULL;
char *w;
int i,j,num_words=0,num_urls=0,first,last;
char wordstr[STRSIZ]="";

typedef struct wrd{
	int  url_id;
	int  count;
} WORD;
WORD * list=NULL;

int cmpword(const void *s1,const void *s2){
	return(
		(((WORD*)s2)->count)-
		(((WORD*)s1)->count)
	);
}

	w=strtok(q,SEPARATOR);wordstr[0]=0;
	while(w){
		num_words++;
		if(*wordstr)sprintf(STREND(wordstr)-1," OR ");
		else sprintf(wordstr,"(");
		sprintf(STREND(wordstr),"dict.word='%s')",w);
		w=strtok(NULL,SEPARATOR);
	}

	if(!num_words)return(0);

	sprintf(qbuf,"SELECT url.rec_id,dict.word FROM dict,url WHERE %s AND url.rec_id=dict.url_id %s%s%s ORDER BY url.rec_id",wordstr,tagstr,statusstr,urlstr);
	res=sql_query(qbuf);if(DbLastErrorCode)return(NULL);
	if(SQL_NUM_ROWS(res)){
		list=(WORD*)malloc(sizeof(WORD)*SQL_NUM_ROWS(res));
		list[0].url_id=0;
		for(i=0;i<SQL_NUM_ROWS(res);i++){
			for(j=0;list[j].url_id;j++){
				if(list[j].url_id==atoi(sql_value(res,i,0))){
					list[j].count++;list[j+1].url_id=0;
					break;
				}
				
			}
			if(!(list[j].url_id)){
				list[j].url_id=atoi(sql_value(res,i,0));
				list[j].count=1;
				list[j+1].url_id=0;
				num_urls++;
			}
		}
		SQL_FREE(res);
		qsort((void*)list,num_urls,sizeof(WORD),cmpword);
		
		/* Exclude results where are not all of the words */
		if(search_mode==MOD_ALL){
			for(i=0;i<num_urls;i++){
				if(list[i].count<num_words){
					num_urls=i;
					break;
				}
			}
		}

		first=np*ps;last=first+ps;
		if(last>num_urls)last=num_urls;
		*found=num_urls;
		num_urls=last-first;
		r=(DOCUMENT*)malloc(sizeof(DOCUMENT)*(num_urls+1));
		for(i=0;i<num_urls;i++){

			sprintf(qbuf,"SELECT \
rec_id,url,content_type,last_modified,title,txt, \
size,last_index_time,next_index_time, \
referrer,keywords,description,crc FROM url \
WHERE rec_id=%d",list[i+first].url_id);
			res=sql_query(qbuf);if(DbLastErrorCode)return(NULL);
			r[i].url_id=list[i+first].url_id;
			r[i].rating=list[i+first].count;
			r[i].url=strdup(sql_value(res,0,1));
			r[i].content_type=strdup(sql_value(res,0,2));
			r[i].last_modified=strdup(sql_value(res,0,3));
			r[i].title=strdup(sql_value(res,0,4));
			r[i].text=strdup(sql_value(res,0,5));
			r[i].size=atoi(sql_value(res,0,6));
			r[i].last_index_time=strdup(sql_value(res,0,7));
			r[i].next_index_time=strdup(sql_value(res,0,8));
			r[i].referrer=atoi(sql_value(res,0,9));
			r[i].keywords=strdup(sql_value(res,0,10));
			r[i].description=strdup(sql_value(res,0,11));
			r[i].crc=strdup(sql_value(res,0,12));
		}
		r[num_urls].url_id=0;
	}else{
		SQL_FREE(res);
	}
	return(r);
}
#else
DOCUMENT * Find(char *q,int np,int ps,int search_mode,int sort_order,char *wordinfo,int *found){
	return(0);
}
#endif
